
from pyroute2.netlink.generic import nlmsg
from pyroute2.netlink.generic import nla


class ndmsg(nlmsg):
    '''
    ARP cache update message

    struct ndmsg {
        unsigned char ndm_family;
        int           ndm_ifindex;  /* Interface index */
        __u16         ndm_state;    /* State */
        __u8          ndm_flags;    /* Flags */
        __u8          ndm_type;
    };

    TODO:
    struct nda_cacheinfo {
        __u32         ndm_confirmed;
        __u32         ndm_used;
        __u32         ndm_updated;
        __u32         ndm_refcnt;
    };
    '''
    fmt = 'BiHBB'

    fields = ('family',
              'ifindex',
              'state',
              'flags',
              'ndm_type')

    # Please note, that nla_map creates implicit
    # enumeration. In this case it will be:
    #
    # NDA_UNSPEC = 0
    # NDA_DST = 1
    # NDA_LLADDR = 2
    # NDA_CACHEINFO = 3
    # NDA_PROBES = 4
    #
    nla_map = (('NDA_UNSPEC', 'hex'),
               ('NDA_DST', 'ipaddr'),
               ('NDA_LLADDR', 'l2addr'),
               ('NDA_CACHEINFO', 'cacheinfo'),
               ('NDA_PROBES', 'uint32'))

    class cacheinfo(nla):
        fmt = 'I' * 4
        fields = ('ndm_confirmed',
                  'ndm_used',
                  'ndm_updated',
                  'ndm_refcnt')



"""
Common utilities
"""


##
# General purpose
#
def map_namespace(prefix, ns):
    """
    Take the namespace prefix, list all constants and build two
    dictionaries -- straight and reverse mappings. E.g.:

    ## neighbor attributes
    NDA_UNSPEC = 0
    NDA_DST = 1
    NDA_LLADDR = 2
    NDA_CACHEINFO = 3
    NDA_PROBES = 4
    (NDA_NAMES, NDA_VALUES) = map_namespace("NDA", globals())

    Will lead to:

    NDA_NAMES = {"NDA_UNSPEC": 0,
                 ...
                 "NDA_PROBES": 4}
    NDA_VALUES = {0: "NDA_UNSPEC",
                  ...
                  4: "NDA_PROBES"}

    """
    by_name = dict([(i, ns[i]) for i in ns.keys() if i.startswith(prefix)])
    by_value = dict([(ns[i], i) for i in ns.keys() if i.startswith(prefix)])
    return (by_name, by_value)


def hexdump(payload, length=0):
    """
    Represent byte string as hex -- for debug purposes
    """
    return ":".join("{0:02x}".format(ord(c))
                    for c in payload[:length] or payload)



import traceback
import threading
import urlparse
import select
import struct
import socket
import Queue
import copy
import time
import os
import io
import uuid
import ssl

from pyroute2.netlink.generic import ctrlmsg
from pyroute2.netlink.generic import nlmsg
from pyroute2.netlink.generic import NETLINK_GENERIC
from pyroute2.netlink.generic import NETLINK_UNUSED


def _monkey_handshake(self):
    ##
    #
    # We have to close incoming connection on handshake error.
    # But if the handshake method is called from the SSLSocket
    # constructor, there is no way to close it: we loose all
    # the references to the failed connection, except the
    # traceback.
    #
    # Using traceback (via sys.exc_info()) can lead to
    # unpredictable consequences with GC. So we have two more
    # choices:
    # 1. use monkey-patch for do_handshake()
    # 2. call it separately.
    #
    # The latter complicates the code by extra checks, that
    # will not be needed most of the time. So the monkey-patch
    # is cheaper.
    #
    ##
    try:
        self._sslobj.do_handshake()
    except Exception as e:
        self._sock.close()
        raise e


ssl.SSLSocket.do_handshake = _monkey_handshake


## Netlink message flags values (nlmsghdr.flags)
#
NLM_F_REQUEST = 1    # It is request message.
NLM_F_MULTI = 2    # Multipart message, terminated by NLMSG_DONE
NLM_F_ACK = 4    # Reply with ack, with zero or error code
NLM_F_ECHO = 8    # Echo this request
# Modifiers to GET request
NLM_F_ROOT = 0x100    # specify tree    root
NLM_F_MATCH = 0x200    # return all matching
NLM_F_ATOMIC = 0x400    # atomic GET
NLM_F_DUMP = (NLM_F_ROOT | NLM_F_MATCH)
# Modifiers to NEW request
NLM_F_REPLACE = 0x100    # Override existing
NLM_F_EXCL = 0x200    # Do not touch, if it exists
NLM_F_CREATE = 0x400    # Create, if it does not exist
NLM_F_APPEND = 0x800    # Add to end of list

NLMSG_NOOP = 0x1    # Nothing
NLMSG_ERROR = 0x2    # Error
NLMSG_DONE = 0x3    # End of a dump
NLMSG_OVERRUN = 0x4    # Data lost
NLMSG_MIN_TYPE = 0x10    # < 0x10: reserved control messages
NLMSG_MAX_LEN = 0xffff  # Max message length

IPRCMD_NOOP = 1
IPRCMD_REGISTER = 2
IPRCMD_UNREGISTER = 3
IPRCMD_STOP = 4
IPRCMD_RELOAD = 5
IPRCMD_ROUTE = 6


class netlink_error(socket.error):
    def __init__(self, code, msg=None):
        msg = msg or os.strerror(code)
        super(netlink_error, self).__init__(code, msg)
        self.code = code


class marshal(object):
    '''
    Generic marshalling class
    '''

    msg_map = {}

    def __init__(self):
        self.lock = threading.Lock()
        # one marshal instance can be used to parse one
        # message at once
        self.buf = None
        self.msg_map = self.msg_map or {}

    def set_buffer(self, init=b''):
        '''
        Set the buffer and return the data length
        '''
        self.buf = io.BytesIO()
        self.buf.write(init)
        self.buf.seek(0)
        return len(init)

    def parse(self, data):
        '''
        Parse the data in the buffer
        '''
        with self.lock:
            total = self.set_buffer(data)
            offset = 0
            result = []

            while offset < total:
                # pick type and length
                (length, msg_type) = struct.unpack('IH', self.buf.read(6))
                error = None
                if msg_type == NLMSG_ERROR:
                    self.buf.seek(16)
                    code = abs(struct.unpack('i', self.buf.read(4))[0])
                    if code > 0:
                        error = netlink_error(code)

                self.buf.seek(offset)
                msg_class = self.msg_map.get(msg_type, nlmsg)
                msg = msg_class(self.buf)
                msg.decode()
                msg['header']['error'] = error
                self.fix_message(msg)
                offset += msg.length
                result.append(msg)

            return result

    def fix_message(self, msg):
        pass


class netlink_socket(socket.socket):
    '''
    Generic netlink socket
    '''

    def __init__(self, family=NETLINK_GENERIC):
        socket.socket.__init__(self, socket.AF_NETLINK,
                               socket.SOCK_DGRAM, family)
        self.pid = os.getpid()
        self.groups = None

    def bind(self, groups=0):
        self.groups = groups
        socket.socket.bind(self, (self.pid, self.groups))


class ssl_credentials(object):
    def __init__(self, key, cert, ca):
        self.key = key
        self.cert = cert
        self.ca = ca


class iothread(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)
        self.setDaemon(True)
        # fd lists for select()
        self._rlist = set()
        self._wlist = set()
        self._xlist = set()
        self._stop = False
        # routing
        self.rtable = {}          # {client_socket: send_method, ...}
        self.families = {}        # {family_id: send_method, ...}
        self.marshals = {}        # {netlink_socket: marshal, ...}
        self.listeners = {}       # {int: Queue(), int: Queue()...}
        self.netlinks = set()     # set(socket, socket...)
        self.clients = set()      # set(socket, socket...)
        self.uplinks = set()      # set(socket, socket...)
        self.servers = set()      # set(socket, socket...)
        self.ssl_keys = {}        # {url: ssl_credentials(), url:...}
        self.mirror = False
        # control socket, for in-process communication only
        self.uuid = uuid.uuid4()
        self.control = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
        self.control.bind(b'\0%s' % (self.uuid))
        self._rlist.add(self.control)
        # debug
        self.record = False
        self.backlog = []

    def _get_socket(self, url, server_side):
        assert type(url) is str
        target = urlparse.urlparse(url)
        hostname = target.hostname or ''
        use_ssl = False
        ssl_version = 2

        if target.scheme[:4] == 'unix':
            if hostname and hostname[0] == '\0':
                address = hostname
            else:
                address = ''.join((hostname, target.path))
            sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        else:
            address = (socket.gethostbyname(hostname), target.port)
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        if target.scheme.find('ssl') >= 0:
            ssl_version = ssl.PROTOCOL_SSLv3
            use_ssl = True

        if target.scheme.find('tls') >= 0:
            ssl_version = ssl.PROTOCOL_TLSv1
            use_ssl = True

        if use_ssl:
            if url not in self.ssl_keys:
                raise Exception('SSL/TLS keys are not provided')

            sock = ssl.wrap_socket(sock,
                                   keyfile=self.ssl_keys[url].key,
                                   certfile=self.ssl_keys[url].cert,
                                   ca_certs=self.ssl_keys[url].ca,
                                   server_side=server_side,
                                   cert_reqs=ssl.CERT_REQUIRED,
                                   ssl_version=ssl_version)
        return (sock, address)

    def route(self, sock, data):
        # message type, offset 4 bytes, length 2 bytes
        mtype = struct.unpack('H', data[4:6])[0]
        # use NETLINK_UNUSED as inter-pyroute2
        # FIXME log routing failures
        if mtype == NETLINK_UNUSED:
            #
            # Control messages
            #
            cmd = self.parse_control(data)
            if cmd['cmd'] == IPRCMD_ROUTE:
                #
                # Route request
                #
                family = cmd.get_attr('CTRL_ATTR_FAMILY_ID')[0]
                if family in self.families:
                    send = self.families[family]
                    self.rtable[sock] = send
                # TODO
                # * subscribe requests
                # * ...

        else:
            #
            # Data messages
            #
            if sock in self.rtable:
                send = self.rtable[sock]
                send(data)

    def parse_control(self, data):
        buf = io.BytesIO()
        buf.write(data)
        buf.seek(0)
        cmd = ctrlmsg(buf)
        cmd.decode()
        return cmd

    def parse(self, data, marshal):
        '''
        Parse and enqueue messages. A message can be
        retrieved from netlink socket as well as from a
        remote system, and it should be properly enqueued
        to make it available for netlink.get() method.

        If iothread.mirror is set, all messages will be also
        copied (mirrored) to the default 0 queue. Please
        make sure that 0 queue exists, before setting
        iothread.mirror to True.

        If there is no such queue for received
        sequence_number, leave sequence_number intact, but
        put the message into default 0 queue, if it exists.
        '''

        for msg in marshal.parse(data):
            if self.record:
                self.backlog.append((time.asctime(), msg))
            key = msg['header']['sequence_number']
            if key not in self.listeners:
                key = 0
            if self.mirror and key != 0:
                self.listeners[0].put(copy.deepcopy(msg))
            if key in self.listeners:
                self.listeners[key].put(msg)

    def command(self, cmd):
        msg = ctrlmsg(io.BytesIO())
        msg['header']['type'] = NETLINK_UNUSED
        msg['cmd'] = cmd
        msg.encode()
        return self.control.sendto(msg.buf.getvalue(),
                                   self.control.getsockname())

    def stop(self):
        return self.command(IPRCMD_STOP)

    def reload(self):
        return self.command(IPRCMD_RELOAD)

    def add_netlink(self, family, groups):
        '''
        Add netlink connection and reload I/O cycle.
        '''
        sock = netlink_socket(family)
        sock.bind(groups)
        self._rlist.add(sock)
        self.netlinks.add(sock)
        self.reload()
        return sock

    def remove_netlink(self, family, groups):
        sock = None
        for sock in self.netlinks:
            if (sock.family, sock.groups) == (family, groups):
                break
        assert sock
        self.netlinks.remove(sock)
        self._rlist.remove(sock)
        sock.close()
        return sock

    def add_server(self, url):
        '''
        Add a server socket to listen for clients on
        '''
        (sock, address) = self._get_socket(url, server_side=False)
        sock.bind(address)
        sock.listen(16)
        self._rlist.add(sock)
        self.servers.add(sock)
        self.reload()
        return sock

    def remove_server(self, sock=None, address=None):
        assert sock or address
        if address:
            for sock in self.servers:
                if sock.getsockname() == address:
                    break
        assert sock
        self._rlist.remove(sock)
        self.servers.remove(sock)
        self.reload()
        return sock

    def add_uplink(self, url):
        '''
        Add an uplink server to get information from
        '''
        (sock, address) = self._get_socket(url, server_side=False)
        sock.connect(address)
        self._rlist.add(sock)
        self.uplinks.add(sock)
        self.reload()
        return sock

    def remove_uplink(self, sock=None, address=None):
        assert sock or address
        if address:
            for sock in self.uplinks:
                if sock.getsockname() == address:
                    break
        assert sock
        self._rlist.remove(sock)
        self.uplinks.remove(sock)
        self.reload()
        return sock

    def add_client(self, sock):
        '''
        Add a client connection. Should not be called
        manually, but only on a client connect.
        '''
        self._rlist.add(sock)
        self._wlist.add(sock)
        self.clients.add(sock)
        self.reload()
        return sock

    def remove_client(self, sock):
        self._rlist.remove(sock)
        self._wlist.remove(sock)
        self.clients.remove(sock)
        return sock

    def run(self):
        while not self._stop:
            [rlist, wlist, xlist] = select.select(self._rlist, [], [])
            for fd in rlist:

                ##
                #
                # Incoming client connections
                #
                if fd in self.servers:
                    try:
                        (client, addr) = fd.accept()
                        self.add_client(client)
                    except ssl.SSLError:
                        # FIXME log SSL errors
                        pass
                    except:
                        traceback.print_exc()
                    continue

                ##
                #
                # Receive data from already open connection
                #
                # FIXME max length
                data = fd.recv(16384)
                if self.record:
                    self.backlog.append((time.asctime(), data))

                ##
                #
                # Control socket, only type == NETLINK_UNUSED
                #
                if fd == self.control:
                    # FIXME move it to specific marshal
                    cmd = self.parse_control(data)
                    if cmd['cmd'] == IPRCMD_STOP:
                        self._stop = True
                        break
                    elif cmd['cmd'] == IPRCMD_RELOAD:
                        pass

                ##
                #
                # Incoming packet from local netlink
                #
                elif fd in self.netlinks:
                    # retranslate to all clients
                    # FIXME what about subscriptions?
                    for sock in self.clients:
                        sock.send(data)
                    # parse message locally
                    self.parse(data, self.marshals[fd])

                ##
                #
                # Incoming packet from remote client
                #
                elif fd in self.clients:
                    if data == '':
                        # client closed connection
                        self.remove_client(fd)
                    else:
                        try:
                            self.route(fd, data)
                        except:
                            # drop packet on any error
                            traceback.print_exc()

                ##
                #
                # Incoming packet from remote uplink
                #
                elif fd in self.uplinks:
                    if data == '':
                        # uplink closed connection
                        self.remove_uplink(fd)
                    else:
                        try:
                            self.parse(data, self.marshals[fd])
                        except:
                            # drop packet on any error
                            pass


class netlink(object):
    '''
    Main netlink messaging class. It automatically spawns threads
    to monitor network and netlink I/O, creates and destroys message
    queues.

    It can operate in three modes:
     * local system only
     * server mode
     * client mode

    By default, it starts in local system only mode. To start a
    server, you should call netlink.serve(url). The method
    can be called several times to listen on specific interfaces
    and/or ports.

    Alternatively, you can start the object in the client mode.
    In that case you should provide server url in the host
    parameter. You can not mix server and client modes, so
    message proxy/relay not possible yet. This will be fixed
    in the future.

    Urls should be specified in the form:
        (host, port)

    E.g.:
        nl = netlink(host=('127.0.0.1', 7000))
    '''

    family = NETLINK_GENERIC
    groups = 0
    marshal = marshal

    def __init__(self, debug=False, host='localsystem', interruptible=False,
                 key=None, cert=None, ca=None):
        self.server = host
        self.iothread = iothread()
        self.listeners = self.iothread.listeners
        self.ssl_keys = self.iothread.ssl_keys
        self.interruptible = interruptible
        if key:
            self.ssl_keys[host] = ssl_credentials(key, cert, ca)
        if host == 'localsystem':
            self.socket = self.iothread.add_netlink(self.family, self.groups)
        else:
            self.socket = self.iothread.add_uplink(self.server)
            self.request_route()
        self.iothread.marshals[self.socket] = self.marshal()
        self.iothread.families[self.family] = self.send
        self.iothread.start()
        self.debug = debug
        self._nonce = 1
        self.servers = {}

    def request_route(self):
        rmsg = ctrlmsg()
        rmsg['header']['type'] = NETLINK_UNUSED
        rmsg['cmd'] = IPRCMD_ROUTE
        rmsg['attrs'] = (('CTRL_ATTR_FAMILY_ID', self.family), )
        rmsg.encode()
        self.socket.send(rmsg.buf.getvalue())

    def shutdown(self, url=None):
        url = url or self.servers.keys()[0]
        self.iothread.remove_server(self.servers[url])
        if url in self.ssl_keys:
            del self.ssl_keys[url]
        del self.servers[url]

    def serve(self, url, key=None, cert=None, ca=None):
        if key:
            self.ssl_keys[url] = ssl_credentials(key, cert, ca)
        self.servers[url] = self.iothread.add_server(url)

    def nonce(self):
        '''
        Increment netlink protocol nonce (there is no need to
        call it directly)
        '''
        if self._nonce == 0xffffffff:
            self._nonce = 1
        else:
            self._nonce += 1
        return self._nonce

    def mirror(self, operate=True):
        '''
        Turn message mirroring on/off. When it is 'on', all
        received messages will be copied (mirrored) into the
        default 0 queue.
        '''
        self.monitor(operate)
        self.iothread.mirror = operate

    def monitor(self, operate=True):
        '''
        Create/destroy the default 0 queue. Netlink socket
        receives messages all the time, and there are many
        messages that are not replies. They are just
        generated by the kernel as a reflection of settings
        changes. To start receiving these messages, call
        netlink.monitor(). They can be fetched by
        netlink.get(0) or just netlink.get().
        '''
        if operate:
            self.listeners[0] = Queue.Queue()
        else:
            del self.listeners[0]

    def get(self, key=0, interruptible=False):
        '''
        Get a message from a queue

        * key -- message queue number
        * interruptible -- catch ctrl-c

        Please note, that setting interruptible=True will cause
        polling overhead. Python starts implied poll cycle, if
        timeout is set.
        '''
        queue = self.listeners[key]
        interruptible = interruptible or self.interruptible
        if interruptible:
            tot = 31536000
        else:
            tot = None
        result = []
        while True:
            # timeout is set to catch ctrl-c
            # Bug-Url: http://bugs.python.org/issue1360
            msg = queue.get(block=True, timeout=tot)
            if msg['header']['error'] is not None:
                raise msg['header']['error']
            if msg['header']['type'] != NLMSG_DONE:
                result.append(msg)
            if (msg['header']['type'] == NLMSG_DONE) or \
               (not msg['header']['flags'] & NLM_F_MULTI):
                break
        # not default queue
        if key != 0:
            # delete the queue
            del self.listeners[key]
            # get remaining messages from the queue and
            # re-route them to queue 0 or drop
            while not queue.empty():
                msg = queue.get(bloc=True, timeout=tot)
                if 0 in self.listeners:
                    self.listeners[0].put(msg)
        return result

    def send(self, buf):
        '''
        Send a buffer or to the local kernel, or to
        the server, depending on the setup.
        '''
        if self.server == 'localsystem':
            self.socket.sendto(buf, (0, 0))
        else:
            self.socket.send(buf)

    def nlm_request(self, msg, msg_type,
                    msg_flags=NLM_F_DUMP | NLM_F_REQUEST):
        '''
        Send netlink request, filling common message
        fields, and wait for response.
        '''
        # FIXME make it thread safe, yeah
        nonce = self.nonce()
        self.listeners[nonce] = Queue.Queue()
        msg['header']['sequence_number'] = nonce
        msg['header']['pid'] = os.getpid()
        msg['header']['type'] = msg_type
        msg['header']['flags'] = msg_flags
        msg.encode()
        self.send(msg.buf.getvalue())
        result = self.get(nonce)
        if not self.debug:
            for i in result:
                del i['header']
        return result



'''
Experimental IP database module. Please, do not use it
unless you understand what are you doing.

Quick start:
    from pyroute2.netlink.ipdb import ipdb
    ip = ipdb()
    ip['eth0'].down()
    ip['eth0']['address'] = '00:11:22:33:44:55'
    ip['eth0']['ifname'] = 'bala'
    ip['eth0']['txqlen'] = 2000
    ip['eth0'].commit()
    ip['bala'].up()
'''
import copy
import threading
from socket import AF_INET
from socket import AF_INET6
from pyroute2.netlink.iproute import iproute
from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg

nla_fields = [i[0] for i in ifinfmsg.nla_map]
nla_fields.append('flags')
nla_fields.append('mask')
nla_fields.append('change')
nla_fields.append('state')


def get_addr_nla(msg):
    nla = None
    if msg['family'] == AF_INET:
        nla = [i[1] for i in msg['attrs']
               if i[0] == 'IFA_LOCAL'][0]
    elif msg['family'] == AF_INET6:
        nla = [i[1] for i in msg['attrs']
               if i[0] == 'IFA_ADDRESS'][0]
    return nla


class upset(set):

    def __init__(self):
        set.__init__(self)
        self.cleanup()
        self.values = {}

    def commit(self):
        for i in self.removed:
            del self.values[i]
        self.cleanup()

    def cleanup(self):
        self.added = set()
        self.removed = set()

    def pop(self, track=True):
        item = set.pop(self)
        if track:
            self.removed.add(item)

    def add(self, key, value=None, track=True):
        if track:
            self.added.add(key)
        self.values[key] = value
        set.add(self, key)

    def remove(self, key, value=None, track=True):
        if track:
            self.removed.add(key)
        # do not remove the value: it will be kept for rollback
        set.remove(self, key)


class dotkeys(dict):

    def __getattribute__(self, key):
        if key in self:
            return self[key]
        else:
            return dict.__getattribute__(self, key)

    def __setattr__(self, key, value):
        if key in self:
            self[key] = value
        else:
            dict.__setattr__(self, key, value)

    def __delattr__(self, key):
        if key in self:
            del self[key]


class interface(dotkeys):
    '''
    Objects of this class represent network interface and
    all related objects:
        * addresses
        * (todo) neighbors
        * (todo) routes

    Interfaces provide transactional model and can act as
    context managers. Any attribute change implicitly
    starts a transaction. The transaction can be managed
    with three methods:
        * review() -- review changes
        * rollback() -- drop all the changes
        * commit() -- try to apply changes

    If anything will go wrong during transaction commit,
    it will be rolled back authomatically and an
    exception will be raised. Failed transaction review
    will be attached to the exception.
    '''
    def __init__(self, dev, ipr=None, parent=None):
        '''
        One can use interface objects standalone as
        well as in connection with ipdb object. Standalone
        usage, though, is discouraged.

        Parameters:
            * dev -- RTM_NEWLINK message
            * ipd -- iproute() reference
            * parent -- ipdb() reference
        '''
        self.ip = ipr
        self.cleanup = ('af_spec',
                        'attrs',
                        'event',
                        'map',
                        'stats',
                        'stats64')
        dev.prefix = 'IFLA_'
        self.last_error = None
        self._parent = parent
        self._fields = [dev.nla2name(i) for i in nla_fields]
        self._updated = {}
        self.load(dev)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            self.commit()
        except Exception as e:
            self.last_error = e

    def load(self, dev):
        '''
        Update the interface info from RTM_NEWLINK message.
        '''
        dev.prefix = 'IFLA_'
        # invalidate old info
        for key in tuple(self.keys()):
            del self[key]
        # load new info
        self.update(dev)
        self._attrs = set()
        for (name, value) in dev['attrs']:
            norm = dev.nla2name(name)
            self._attrs.add(norm)
            self[norm] = value
        for item in self.cleanup:
            if item in self:
                del self[item]
        self.old_name = self['ifname']
        self._updated = {}
        if self._parent is not None:
            self['ipaddr'] = self._parent.ipaddr[self['index']]

    def __setitem__(self, key, value):
        if key in self:
            self._updated[key] = self[key]
        dotkeys.__setitem__(self, key, value)

    def review(self):
        '''
        Review the changes that are not commited yet. Output
        format can be changed later.
        '''
        response = {'ipaddr': [],
                    'attrs': {}}
        for i in self['ipaddr'].added:
            response['ipaddr'].append('+%s/%s' % tuple(i))
        for i in self['ipaddr'].removed:
            response['ipaddr'].append('-%s/%s' % tuple(i))
        for i in tuple(self._updated.keys()):
            response['attrs'][i] = '%s -> %s' % (self._updated[i],
                                                 self[i])
        return response

    def rollback(self):
        '''
        Drop all not applied changes and rollback transaction.
        '''
        for i in self['ipaddr'].added:
            self['ipaddr'].remove(i, track=False)
        for i in self['ipaddr'].removed:
            self['ipaddr'].add(i, track=False)
        self['ipaddr'].cleanup()
        for i in tuple(self._updated.keys()):
            self[i] = self._updated[i]
            del self._updated[i]

    def commit(self):
        '''
        Commit transaction. In the case of exception all
        changes applied during commit will be reverted.
        '''
        e = None
        transaction = copy.deepcopy(self.review())
        try:
            # commit IP address changes
            for i in self['ipaddr'].added:
                # add address
                self.ip.addr_add(self['index'], i[0], i[1])
            for i in self['ipaddr'].removed:
                self.ip.addr_del(self['index'], i[0], i[1])
            # commit interface changes
            request = {}
            for key in self._updated:
                if key in self._fields:
                    request[key] = self[key]
            self.ip.link('set', self['index'], **request)
            # flush IP address changes
            self['ipaddr'].commit()
        except Exception as e:
            # something went wrong: roll the transaction back
            try:
                # remove all added and add all removed
                for i in self['ipaddr'].added:
                    self.ip.addr_del(self['index'], i[0], i[1])
                for i in self['ipaddr'].removed:
                    self.ip.addr_add(self['index'], i[0], i[1])
                # apply old attributes
                request = {}
                for key in tuple(self._updated.keys()):
                    if key in self._fields:
                        request[key] = self._updated[key]
                        del self._updated[key]
                self.ip.link('set', self['index'], **request)
                # rollback IP address tracking
                self['ipaddr'].cleanup()
            except:
                pass
        # re-load interface information
        self.load(self.ip.get_links(self['index'])[0])
        if e is not None:
            e.transaction = transaction
            raise e

    def up(self):
        '''
        Shortcut: change interface state to 'up'.

        Requres commit.
        '''
        self['flags'] |= 1

    def down(self):
        '''
        Shortcut: change interface state to 'down'.

        Requires commit.
        '''
        self['flags'] &= ~(self['flags'] & 1)


class ipdb(dotkeys):
    '''
    The class that maintains information about network setup
    of the host. Monitoring netlink events allows it to react
    immediately. It uses no polling.

    No methods of the class should be called directly.
    '''

    def __init__(self, ipr=None, host='localsystem',
                 key=None, cert=None, ca=None):
        '''
        Parameters:
            * ipr -- iproute() reference

        If you do not provide iproute instance, ipdb will
        start it automatically. Please note, that there can
        be only one iproute instance per process. Actually,
        you can start two and more iproute instances, but
        only the first one will receive anything.
        '''
        self.ip = ipr or iproute(host=host, key=key, cert=cert, ca=ca)

        # caches
        self.ipaddr = {}
        self.routes = {}
        self.neighbors = {}

        # load information on startup
        self.update_links(self.ip.get_links())
        self.update_addr(self.ip.get_addr())

        # start monitoring thread
        self.ip.monitor()
        self.ip.mirror()
        self._mthread = threading.Thread(target=self.monitor)
        self._mthread.setDaemon(True)
        self._mthread.start()

    def update_links(self, links):
        '''
        Rebuild links index from list of RTM_NEWLINK messages.
        '''
        for dev in links:
            if dev['index'] not in self.ipaddr:
                self.ipaddr[dev['index']] = upset()
            i = self[dev['index']] = interface(dev, self.ip, self)
            self[i['ifname']] = i

    def update_addr(self, addrs, action='add'):
        '''
        Update interface list of an interface.
        '''
        for addr in addrs:
            nla = get_addr_nla(addr)
            if nla is not None:
                method = getattr(self.ipaddr[addr['index']], action)
                try:
                    method(key=(nla, addr['prefixlen']),
                           value=addr, track=False)
                except:
                    pass

    def monitor(self):
        '''
        Main monitoring cycle. It gets messages from the
        default iproute queue and updates objects in the
        database.
        '''
        while True:
            try:
                messages = self.ip.get()
            except:
                continue
            for msg in messages:
                if msg.get('event', None) == 'RTM_NEWLINK':
                    index = msg['index']
                    if index in self:
                        # get old name
                        old_name = self[index].old_name
                        # load message
                        self[index].load(msg)
                        # check for new name
                        if self[index]['ifname'] != old_name:
                            del self[old_name]
                            self[self[index]['ifname']] = self[index]
                    else:
                        self.update_links([msg])
                elif msg.get('event', None) == 'RTM_DELLINK':
                    del self[self[msg['index']]['ifname']]
                    del self[msg['index']]
                elif msg.get('event', None) == 'RTM_NEWADDR':
                    self.update_addr([msg], 'add')
                elif msg.get('event', None) == 'RTM_DELADDR':
                    self.update_addr([msg], 'remove')




from pyroute2.netlink import netlink
from pyroute2.netlink import NLM_F_REQUEST
from pyroute2.netlink import marshal
from pyroute2.netlink.generic import nla
from pyroute2.netlink.generic import genlmsg
from pyroute2.netlink.generic import ctrlmsg

TASKSTATS_CMD_UNSPEC = 0      # Reserved
TASKSTATS_CMD_GET = 1         # user->kernel request/get-response
TASKSTATS_CMD_NEW = 2

NLMSG_MIN_TYPE = 0x10

GENL_NAMSIZ = 16    # length of family name
GENL_MIN_ID = NLMSG_MIN_TYPE
GENL_MAX_ID = 1023

GENL_ADMIN_PERM = 0x01
GENL_CMD_CAP_DO = 0x02
GENL_CMD_CAP_DUMP = 0x04
GENL_CMD_CAP_HASPOL = 0x08

#
# List of reserved static generic netlink identifiers:
#
GENL_ID_GENERATE = 0
GENL_ID_CTRL = NLMSG_MIN_TYPE

#
# Controller
#

CTRL_CMD_UNSPEC = 0x0
CTRL_CMD_NEWFAMILY = 0x1
CTRL_CMD_DELFAMILY = 0x2
CTRL_CMD_GETFAMILY = 0x3
CTRL_CMD_NEWOPS = 0x4
CTRL_CMD_DELOPS = 0x5
CTRL_CMD_GETOPS = 0x6
CTRL_CMD_NEWMCAST_GRP = 0x7
CTRL_CMD_DELMCAST_GRP = 0x8
CTRL_CMD_GETMCAST_GRP = 0x9  # unused


CTRL_ATTR_UNSPEC = 0x0
CTRL_ATTR_FAMILY_ID = 0x1
CTRL_ATTR_FAMILY_NAME = 0x2
CTRL_ATTR_VERSION = 0x3
CTRL_ATTR_HDRSIZE = 0x4
CTRL_ATTR_MAXATTR = 0x5
CTRL_ATTR_OPS = 0x6
CTRL_ATTR_MCAST_GROUPS = 0x7

CTRL_ATTR_OP_UNSPEC = 0x0
CTRL_ATTR_OP_ID = 0x1
CTRL_ATTR_OP_FLAGS = 0x2

CTRL_ATTR_MCAST_GRP_UNSPEC = 0x0
CTRL_ATTR_MCAST_GRP_NAME = 0x1
CTRL_ATTR_MCAST_GRP_ID = 0x2


class tcmd(genlmsg):
    nla_map = (('TASKSTATS_CMD_ATTR_UNSPEC', 'none'),
               ('TASKSTATS_CMD_ATTR_PID', 'uint32'),
               ('TASKSTATS_CMD_ATTR_TGID', 'uint32'),
               ('TASKSTATS_CMD_ATTR_REGISTER_CPUMASK', 'uint32'),
               ('TASKSTATS_CMD_ATTR_DEREGISTER_CPUMASK', 'uint32'))


class tstats(nla):
    t_fields = (('version', 'H'),                           # 2
                ('ac_exitcode', 'I'),                       # 4
                ('ac_flag', 'B'),                           # 1
                ('ac_nice', 'B'),                           # 1 --- 8
                ('cpu_count', 'Q'),                         # 8
                ('cpu_delay_total', 'Q'),                   # 8
                ('blkio_count', 'Q'),                       # 8
                ('blkio_delay_total', 'Q'),                 # 8
                ('swapin_count', 'Q'),                      # 8
                ('swapin_delay_total', 'Q'),                # 8
                ('cpu_run_real_total', 'Q'),                # 8
                ('cpu_run_virtual_total', 'Q'),             # 8
                ('ac_comm', '32s'),                         # 32 +++ 112
                ('ac_sched', 'B'),                          # 1
                ('__pad', '5x'),                            # 5 --- 8 (!)
                ('ac_uid', 'I'),                            # 4  +++ 120
                ('ac_gid', 'I'),                            # 4
                ('ac_pid', 'I'),                            # 4
                ('ac_ppid', 'I'),                           # 4
                ('ac_btime', 'I'),                          # 4  +++ 136
                ('ac_etime', 'Q'),                          # 8  +++ 144
                ('ac_utime', 'Q'),                          # 8
                ('ac_stime', 'Q'),                          # 8
                ('ac_minflt', 'Q'),                         # 8
                ('ac_majflt', 'Q'),                         # 8
                ('coremem', 'Q'),                           # 8
                ('virtmem', 'Q'),                           # 8
                ('hiwater_rss', 'Q'),                       # 8
                ('hiwater_vm', 'Q'),                        # 8
                ('read_char', 'Q'),                         # 8
                ('write_char', 'Q'),                        # 8
                ('read_syscalls', 'Q'),                     # 8
                ('write_syscalls', 'Q'),                    # 8
                ('read_bytes', 'Q'),                        # ...
                ('write_bytes', 'Q'),
                ('cancelled_write_bytes', 'Q'),
                ('nvcsw', 'Q'),
                ('nivcsw', 'Q'),
                ('ac_utimescaled', 'Q'),
                ('ac_stimescaled', 'Q'),
                ('cpu_scaled_run_real_total', 'Q'))

    def decode(self):
        nla.decode(self)
        self['ac_comm'] = self['ac_comm'][:self['ac_comm'].find('\0')]


class taskstatsmsg(genlmsg):

    nla_map = (('TASKSTATS_TYPE_UNSPEC', 'none'),
               ('TASKSTATS_TYPE_PID', 'uint32'),
               ('TASKSTATS_TYPE_TGID', 'uint32'),
               ('TASKSTATS_TYPE_STATS', 'stats'),
               ('TASKSTATS_TYPE_AGGR_PID', 'aggr_pid'),
               ('TASKSTATS_TYPE_AGGR_TGID', 'aggr_tgid'))

    class stats(tstats):
        pass  # FIXME: optimize me!

    class aggr_pid(nla):
        nla_map = (('TASKSTATS_TYPE_UNSPEC', 'none'),
                   ('TASKSTATS_TYPE_PID', 'uint32'),
                   ('TASKSTATS_TYPE_TGID', 'uint32'),
                   ('TASKSTATS_TYPE_STATS', 'stats'))

        class stats(tstats):
            pass

    class aggr_tgid(nla):
        nla_map = (('TASKSTATS_TYPE_UNSPEC', 'none'),
                   ('TASKSTATS_TYPE_PID', 'uint32'),
                   ('TASKSTATS_TYPE_TGID', 'uint32'),
                   ('TASKSTATS_TYPE_STATS', 'stats'))

        class stats(tstats):
            pass


class taskstats(netlink):

    marshal = marshal

    def __init__(self):
        netlink.__init__(self)
        # FIXME
        self.io_thread.marshal.msg_map[GENL_ID_CTRL] = ctrlmsg
        self.prid = self.get_protocol_id('TASKSTATS')
        self.io_thread.marshal.msg_map[self.prid] = taskstatsmsg

    def get_protocol_id(self, prid):
        msg = ctrlmsg()
        msg['cmd'] = CTRL_CMD_GETFAMILY
        msg['version'] = 1
        msg['attrs'].append(('CTRL_ATTR_FAMILY_NAME', prid))
        response = self.nlm_request(msg, GENL_ID_CTRL,
                                    msg_flags=NLM_F_REQUEST)[0]
        prid = [i[1] for i in response['attrs']
                if i[0] == 'CTRL_ATTR_FAMILY_ID'][0]
        return prid

    def get_pid_stat(self, pid):
        msg = tcmd()
        msg['cmd'] = TASKSTATS_CMD_GET
        msg['version'] = 1
        msg['attrs'].append(('TASKSTATS_CMD_ATTR_PID', pid))
        return self.nlm_request(msg, self.prid, msg_flags=NLM_F_REQUEST)

    def get_mask_stat(self, mask):
        pass




from pyroute2.netlink.generic import nlmsg
from pyroute2.netlink.generic import nla


class rtmsg(nlmsg):
    '''
    Routing update message

    struct rtmsg {
        unsigned char rtm_family;   /* Address family of route */
        unsigned char rtm_dst_len;  /* Length of destination */
        unsigned char rtm_src_len;  /* Length of source */
        unsigned char rtm_tos;      /* TOS filter */

        unsigned char rtm_table;    /* Routing table ID */
        unsigned char rtm_protocol; /* Routing protocol; see below */
        unsigned char rtm_scope;    /* See below */
        unsigned char rtm_type;     /* See below */

        unsigned int  rtm_flags;
    };
    '''
    fmt = 'BBBBBBBBI'

    fields = ('family',
              'dst_len',
              'src_len',
              'tos',
              'table',
              'proto',
              'scope',
              'type',
              'flags')

    nla_map = (('RTA_UNSPEC', 'none'),
               ('RTA_DST', 'ipaddr'),
               ('RTA_SRC', 'ipaddr'),
               ('RTA_IIF', 'uint32'),
               ('RTA_OIF', 'uint32'),
               ('RTA_GATEWAY', 'ipaddr'),
               ('RTA_PRIORITY', 'uint32'),
               ('RTA_PREFSRC', 'ipaddr'),
               ('RTA_METRICS', 'uint32'),
               ('RTA_MULTIPATH', 'hex'),
               ('RTA_PROTOINFO', 'hex'),
               ('RTA_FLOW', 'hex'),
               ('RTA_CACHEINFO', 'cacheinfo'),
               ('RTA_SESSION', 'hex'),
               ('RTA_MP_ALGO', 'hex'),
               ('RTA_TABLE', 'uint32'))

    class cacheinfo(nla):
        fmt = 'IIiIIIII'
        fields = ('rta_clntref',
                  'rta_lastuse',
                  'rta_expires',
                  'rta_error',
                  'rta_used',
                  'rta_id',
                  'rta_ts',
                  'rta_tsage')




import socket
import struct
import io

from pyroute2.common import hexdump

##  Netlink family
#
NETLINK_ROUTE = 0            # Routing/device hook
NETLINK_UNUSED = 1           # Unused number
NETLINK_USERSOCK = 2         # Reserved for user mode socket protocols
NETLINK_FIREWALL = 3         # Firewalling hook
NETLINK_INET_DIAG = 4        # INET socket monitoring
NETLINK_NFLOG = 5            # netfilter/iptables ULOG
NETLINK_XFRM = 6             # ipsec
NETLINK_SELINUX = 7          # SELinux event notifications
NETLINK_ISCSI = 8            # Open-iSCSI
NETLINK_AUDIT = 9            # auditing
NETLINK_FIB_LOOKUP = 10
NETLINK_CONNECTOR = 11
NETLINK_NETFILTER = 12       # netfilter subsystem
NETLINK_IP6_FW = 13
NETLINK_DNRTMSG = 14         # DECnet routing messages
NETLINK_KOBJECT_UEVENT = 15  # Kernel messages to userspace
NETLINK_GENERIC = 16
# leave room for NETLINK_DM (DM Events)
NETLINK_SCSITRANSPORT = 18   # SCSI Transports


NLMSG_ALIGNTO = 4


def NLMSG_ALIGN(l):
    return (l + NLMSG_ALIGNTO - 1) & ~ (NLMSG_ALIGNTO - 1)


class NotInitialized(Exception):
    pass


class nlmsg_base(dict):
    '''
    Netlink base class. You do not need to inherit it directly, unless
    you're inventing completely new protocol structure.

    Use nlmsg or nla classes.

    ...

    Netlink message structure

    | header | data |

    nlmsg header:
        + uint32 length
        + uint16 type
        + uint16 flags
        + uint32 sequence number
        + uint32 pid

    nla header:
        + uint16 length
        + uint16 type

    data:
        + data-specific struct
        + NLA
        + NLA
        + ...

    To describe data, you can use one of the ways:

    1. 'fmt' and 'fields' class atributes, e.g.:

        fmt = 'HH'
        fields = ('length',
                  'type')

        this will decode/encode two 16bit unsigned integers
        into/from fields 'length' and 'type'.

    2. 't_fields' attribute, e.g.:

        t_fields = (('length', 'H'),
                    ('type', 'H'))

        will do the same.

    NLAs are decoded/encoded according to 'nla_map':

        nla_map = (('NDA_UNSPEC', 'none'),
                   ('NDA_DST', 'ipaddr'),
                   ('NDA_LLADDR', 'l2addr'),
                   ('NDA_CACHEINFO', 'cacheinfo'),
                   ('NDA_PROBES', 'uint32'))

    Please note, that 'nla_map' creates implied enumeration from
    its fields. In the example above NDA_UNSPEC == 0 and
    NDA_PROBES == 4. These numbers will be used as uint16 'type'
    in NLA header.

    List of public types as 'none', 'uint32', 'ipaddr', 'asciiz'
    you can read in nlmsg_atoms class.
    '''

    fmt = ''                    # data format string, see struct
    fields = ('value', )        # data field names, to build a dictionary
    t_fields = NotInitialized
    header = None               # optional header class
    nla_map = {}                # NLA mapping

    def __init__(self, buf=None, length=None, parent=None):
        dict.__init__(self)
        for i in self.fields:
            self[i] = 0  # FIXME: only for number values
        self.buf = buf or io.BytesIO()
        self.length = length or 0
        self.parent = parent
        self.offset = 0
        self.prefix = None
        self['attrs'] = []
        self['value'] = NotInitialized
        self.value = NotInitialized
        if self.header is not None:
            self['header'] = self.header(self.buf)
        self.register_fields()
        self.register_nlas()

    def nla2name(self, name):
        '''
        Convert NLA name into human-friendly name

        Example: IFLA_ADDRESS -> address

        Requires self.prefix to be set
        '''
        return name[(name.find(self.prefix) + 1) * len(self.prefix):].lower()

    def name2nla(self, name):
        '''
        Convert human-friendly name into NLA name

        Example: address -> IFLA_ADDRESS

        Requires self.prefix to be set
        '''
        name = name.upper()
        if name.find(self.prefix) == -1:
            name = "%s%s" % (self.prefix, name)
        return name

    def reserve(self):
        '''
        Reserve space in the buffer for data. This can be used
        to skip encoding of the header until some fields will
        be known.
        '''
        self.buf.seek(struct.calcsize(self.fmt), 1)

    def decode(self):
        self.offset = self.buf.tell()
        # read the header
        if self.header is not None:
            self['header'].decode()
            # update length from header
            # it can not be less than 4
            self.length = max(self['header']['length'], 4)
            if self.fmt == 's':
                self.fmt = '%is' % (self.length - 4)
            elif self.fmt == 'z':
                self.fmt = '%is' % (self.length - 5)
        # read the data
        size = struct.calcsize(self.fmt)
        self.update(dict(zip(self.fields,
                         struct.unpack(self.fmt, self.buf.read(size)))))
        # align NLA chain start
        self.buf.seek(NLMSG_ALIGN(self.buf.tell()))
        # read NLA chain
        self.decode_nlas()
        if len(self['attrs']) == 0:
            del self['attrs']
        if self['value'] is NotInitialized:
            del self['value']

    def encode(self):
        init = self.buf.tell()
        # reserve space for the header
        if self.header is not None:
            self['header'].reserve()
        if self.fmt == 's':
            length = len(self.fields[0]) + 4
            self.fmt = '%is' % (length)
        elif self.fmt == 'z':
            length = len(self.fields[0]) + 5
            self.fmt = '%is' % (length)
        payload = struct.pack(self.fmt, *([self[i] for i in self.fields]))
        diff = NLMSG_ALIGN(len(payload)) - len(payload)
        self.buf.write(payload)
        self.buf.write(b'\0' * diff)
        # write NLA chain
        self.encode_nlas()
        # calculate the size and write it
        if self.header is not None:
            save = self.buf.tell()
            self['header']['length'] = save - init
            self.buf.seek(init)
            self['header'].encode()
            self.buf.seek(save)

    def setvalue(self, value):
        if type(value) is dict:
            self.update(value)
        else:
            self['value'] = value
            self.value = value

    def get_attr(self, attr):
        '''
        Return attr by name
        '''
        return [i[1] for i in self['attrs'] if i[0] == attr]

    def getvalue(self):
        '''
        Atomic NLAs return their value in the 'value' field,
        not as a dictionary. Complex NLAs return whole dictionary.
        '''
        if self.value != NotInitialized:
            # value decoded by custom decoder
            return self.value
        elif 'value' in self:
            # raw value got by generic decoder
            return self['value']
        else:
            return self

    def register_fields(self):
        '''
        Convert 't_fields' attribute into 'fmt' and 'fields'.
        '''
        if self.t_fields is NotInitialized:
            return

        fields = []
        fmt = []

        for i in self.t_fields:
            fmt.append(i[1])
            if (i[1].find('x') == -1) and (i[0] != '__pad'):
                fields.append(i[0])

        self.fmt = ''.join(fmt)
        self.fields = tuple(fields)

    def register_nlas(self):
        '''
        Convert 'nla_map' tuple into two dictionaries for mapping
        and reverse mapping of NLA types.
        '''
        # clean up NLA mappings
        self.t_nla_map = {}
        self.r_nla_map = {}

        # create enumeration
        types = enumerate((i[0] for i in self.nla_map))
        # that's a little bit tricky, but to reduce
        # the required amount of code in modules, we have
        # to jump over the head
        zipped = [(k[1][0], k[0][0], k[0][1]) for k in
                  zip(self.nla_map, types)]

        for (key, name, nla_class) in zipped:
            # lookup NLA class
            nla_class = getattr(self, nla_class)
            # update mappings
            self.t_nla_map[key] = (nla_class, name)
            self.r_nla_map[name] = (nla_class, key)

    def encode_nlas(self):
        for i in self['attrs']:
            if i[0] in self.r_nla_map:
                msg_class = self.r_nla_map[i[0]][0]
                msg_type = self.r_nla_map[i[0]][1]
                try:
                    # encode NLA
                    nla = msg_class(self.buf, parent=self)
                    nla['header']['type'] = msg_type
                    nla.setvalue(i[1])
                    nla.encode()
                except:
                    # FIXME
                    import traceback
                    traceback.print_exc()

    def decode_nlas(self):
        while self.buf.tell() < (self.offset + self.length):
            init = self.buf.tell()
            # pick the length and the type
            (length, msg_type) = struct.unpack('HH', self.buf.read(4))
            # rewind to the beginning
            self.buf.seek(init)
            length = min(max(length, 4),
                         (self.length - self.buf.tell() + self.offset))

            # we have a mapping for this NLA
            if msg_type in self.t_nla_map:
                # get the class
                msg_class = self.t_nla_map[msg_type][0]
                # and the name
                msg_name = self.t_nla_map[msg_type][1]
                try:
                    # decode NLA
                    nla = msg_class(self.buf, length, self)
                    nla.decode()
                    self['attrs'].append((msg_name, nla.getvalue()))
                except:
                    # FIXME
                    import traceback
                    traceback.print_exc()
                    self.buf.seek(init)
                    self['attrs'].append((msg_name,
                                          hexdump(self.buf.read(length))))

            # fix the offset
            self.buf.seek(init + NLMSG_ALIGN(length))


class nla_header(nlmsg_base):
    fmt = 'HH'
    fields = ('length', 'type')


class nla_base(nlmsg_base):
    header = nla_header


class nlmsg_header(nlmsg_base):
    '''
    Common netlink message header
    '''
    fmt = 'IHHII'
    fields = ('length', 'type', 'flags', 'sequence_number', 'pid')


class nlmsg_atoms(nlmsg_base):
    '''
    A collection of base NLA types
    '''
    class none(nla_base):
        '''
        'none' type is used to skip decoding of NLA. You can
        also use 'hex' type to dump NLA's content.
        '''
        def decode(self):
            nla_base.decode(self)
            self.value = None

    class uint8(nla_base):
        fmt = '=B'

    class uint16(nla_base):
        fmt = '=H'

    class uint32(nla_base):
        fmt = '=I'

    class ipaddr(nla_base):
        '''
        This class is used to decode IP addresses according to
        the family. Socket library currently supports only two
        families, AF_INET and AF_INET6.

        We do not specify here the string size, it will be
        calculated in runtime.
        '''
        fmt = 's'

        def encode(self):
            self['value'] = socket.inet_pton(self.parent['family'], self.value)
            nla_base.encode(self)

        def decode(self):
            nla_base.decode(self)
            self.value = socket.inet_ntop(self.parent['family'], self['value'])

    class l2addr(nla_base):
        '''
        Decode MAC address.
        '''
        fmt = '=6s'

        def encode(self):
            self['value'] = ''.join((chr(int(i, 16)) for i in
                                     self.value.split(':')))
            nla_base.encode(self)

        def decode(self):
            nla_base.decode(self)
            self.value = ':'.join('%02x' % (ord(i)) for i in self['value'])

    class hex(nla_base):
        '''
        Represent NLA's content with header as hex string.
        '''
        fmt = 's'

        def decode(self):
            nla_base.decode(self)
            self.value = hexdump(self['value'])

    class asciiz(nla_base):
        '''
        Zero-terminated string.
        '''
        # FIXME: move z-string hacks from general decode here?
        fmt = 'z'


class nla(nla_base, nlmsg_atoms):
    '''
    Main NLA class
    '''
    def decode(self):
        nla_base.decode(self)
        del self['header']


class nlmsg(nlmsg_atoms):
    '''
    Main netlink message class
    '''
    header = nlmsg_header


class genlmsg(nlmsg):
    '''
    Generic netlink message
    '''
    fmt = 'BBH'
    fields = ('cmd',
              'version',
              'reserved')


class ctrlmsg(genlmsg):
    '''
    Netlink control message
    '''
    # FIXME: to be extended
    nla_map = (('CTRL_ATTR_UNSPEC', 'none'),
               ('CTRL_ATTR_FAMILY_ID', 'uint16'),
               ('CTRL_ATTR_FAMILY_NAME', 'asciiz'))




from pyroute2.netlink.generic import nlmsg
from pyroute2.netlink.generic import nla

## address attributes
#
# Important comment:
# IFA_ADDRESS is prefix address, rather than local interface address.
# It makes no difference for normally configured broadcast interfaces,
# but for point-to-point IFA_ADDRESS is DESTINATION address,
# local address is supplied in IFA_LOCAL attribute.
#


class ifaddrmsg(nlmsg):
    '''
    IP address information

    struct ifaddrmsg {
       unsigned char ifa_family;    /* Address type */
       unsigned char ifa_prefixlen; /* Prefixlength of address */
       unsigned char ifa_flags;     /* Address flags */
       unsigned char ifa_scope;     /* Address scope */
       int           ifa_index;     /* Interface index */
    };

    '''
    fmt = 'BBBBI'
    fields = ('family',
              'prefixlen',
              'flags',
              'scope',
              'index')

    nla_map = (('IFA_UNSPEC',  'hex'),
               ('IFA_ADDRESS', 'ipaddr'),
               ('IFA_LOCAL', 'ipaddr'),
               ('IFA_LABEL', 'asciiz'),
               ('IFA_BROADCAST', 'ipaddr'),
               ('IFA_ANYCAST', 'ipaddr'),
               ('IFA_CACHEINFO', 'cacheinfo'),
               ('IFA_MULTICAST', 'ipaddr'))

    class cacheinfo(nla):
        fmt = 'I' * 4
        fields = ('ifa_prefered',
                  'ifa_valid',
                  'cstamp',
                  'tstamp')




from pyroute2.netlink.iproute import iproute
from pyroute2.netlink.ipdb import ipdb
from pyroute2.netlink.taskstats import taskstats

make_pep8_happy = iproute
make_pep8_happy = ipdb
make_pep8_happy = taskstats



from pyroute2.netlink.generic import nla
from pyroute2.netlink.generic import nlmsg

states = ('UNKNOWN',
          'NOTPRESENT',
          'DOWN',
          'LOWERLAYERDOWN',
          'TESTING',
          'DORMANT',
          'UP')
state_by_name = dict(((i[1], i[0]) for i in enumerate(states)))
state_by_code = dict(enumerate(states))


class ifinfmsg(nlmsg):
    '''
    Network interface message
    struct ifinfomsg {
        unsigned char  ifi_family; /* AF_UNSPEC */
        unsigned short ifi_type;   /* Device type */
        int            ifi_index;  /* Interface index */
        unsigned int   ifi_flags;  /* Device flags  */
        unsigned int   ifi_change; /* change mask */
    };
    '''
    fmt = 'BHiII'
    fields = ('family',
              'ifi_type',
              'index',
              'flags',
              'change')

    nla_map = (('IFLA_UNSPEC', 'none'),
               ('IFLA_ADDRESS', 'l2addr'),
               ('IFLA_BROADCAST', 'l2addr'),
               ('IFLA_IFNAME', 'asciiz'),
               ('IFLA_MTU', 'uint32'),
               ('IFLA_LINK', 'uint32'),
               ('IFLA_QDISC', 'asciiz'),
               ('IFLA_STATS', 'ifstats'),
               ('IFLA_COST', 'hex'),
               ('IFLA_PRIORITY', 'hex'),
               ('IFLA_MASTER', 'uint32'),
               ('IFLA_WIRELESS', 'hex'),
               ('IFLA_PROTINFO', 'hex'),
               ('IFLA_TXQLEN', 'uint32'),
               ('IFLA_MAP', 'ifmap'),
               ('IFLA_WEIGHT', 'hex'),
               ('IFLA_OPERSTATE', 'state'),
               ('IFLA_LINKMODE', 'uint8'),
               ('IFLA_LINKINFO', 'ifinfo'),
               ('IFLA_NET_NS_PID', 'hex'),
               ('IFLA_IFALIAS', 'hex'),
               ('IFLA_NUM_VF', 'uint32'),
               ('IFLA_VFINFO_LIST', 'hex'),
               ('IFLA_STATS64', 'ifstats64'),
               ('IFLA_VF_PORTS', 'hex'),
               ('IFLA_PORT_SELF', 'hex'),
               ('IFLA_AF_SPEC', 'af_spec'),
               ('IFLA_GROUP', 'uint32'),
               ('IFLA_NET_NS_FD', 'hex'),
               ('IFLA_EXT_MASK', 'hex'),
               ('IFLA_PROMISCUITY', 'uint32'),
               ('IFLA_NUM_TX_QUEUES', 'uint32'),
               ('IFLA_NUM_RX_QUEUES', 'uint32'))

    class state(nla):
        fmt = '=B'

        def encode(self):
            self['value'] = state_by_name[self.value]
            nla.encode(self)

        def decode(self):
            nla.decode(self)
            self.value = state_by_code[self['value']]

    class ifstats(nla):
        fmt = 'I' * 23
        fields = ('rx_packets',
                  'tx_packets',
                  'rx_bytes',
                  'tx_bytes',
                  'rx_errors',
                  'tx_errors',
                  'rx_dropped',
                  'tx_dropped',
                  'multicast',
                  'collisions',
                  'rx_length_errors',
                  'rx_over_errors',
                  'rx_crc_errors',
                  'rx_frame_errors',
                  'rx_fifo_errors',
                  'rx_missed_errors',
                  'tx_aborted_errors',
                  'tx_carrier_errors',
                  'tx_fifo_errors',
                  'tx_heartbeat_errors',
                  'tx_window_errors',
                  'rx_compressed',
                  'tx_compressed')

    class ifstats64(ifstats):
        fmt = 'Q' * 23

    class ifmap(nla):
        fmt = 'QQQHBB'
        fields = ('mem_start', 'mem_end', 'base_addr', 'irq', 'dma', 'port')

    class ifinfo(nla):
        nla_map = (('IFLA_INFO_UNSPEC', 'none'),
                   ('IFLA_INFO_KIND', 'asciiz'),
                   ('IFLA_INFO_DATA', 'hex'),
                   ('IFLA_INFO_XSTATS', 'hex'))

    class af_spec(nla):
        nla_map = (('AF_UNSPEC', 'none'),
                   ('AF_UNIX', 'hex'),
                   ('AF_INET', 'inet'),
                   ('AF_AX25', 'hex'),
                   ('AF_IPX', 'hex'),
                   ('AF_APPLETALK', 'hex'),
                   ('AF_NETROM', 'hex'),
                   ('AF_BRIDGE', 'hex'),
                   ('AF_ATMPVC', 'hex'),
                   ('AF_X25', 'hex'),
                   ('AF_INET6', 'inet6'))

        class inet(nla):
            fmt = 'I' * 26
            #  ./include/linux/inetdevice.h: struct ipv4_devconf
            fields = ('sysctl',
                      'forwarding',
                      'mc_forwarding',
                      'proxy_arp',
                      'accept_redirects',
                      'secure_redirects',
                      'send_redirects',
                      'shared_media',
                      'rp_filter',
                      'accept_source_route',
                      'bootp_relay',
                      'log_martians',
                      'tag',
                      'arp_filter',
                      'medium_id',
                      'disable_xfrm',
                      'disable_policy',
                      'force_igmp_version',
                      'arp_announce',
                      'arp_ignore',
                      'promote_secondaries',
                      'arp_accept',
                      'arp_notify',
                      'accept_local',
                      'src_valid_mark',
                      'proxy_arp_pvlan',
                      'route_localnet')

        class inet6(nla):
            nla_map = (('IFLA_INET6_UNSPEC', 'none'),
                       ('IFLA_INET6_FLAGS', 'uint32'),
                       ('IFLA_INET6_CONF', 'ipv6_devconf'),
                       ('IFLA_INET6_STATS', 'ipv6_stats'),
                       ('IFLA_INET6_MCAST', 'hex'),
                       ('IFLA_INET6_CACHEINFO', 'ipv6_cache_info'),
                       ('IFLA_INET6_ICMP6STATS', 'icmp6_stats'))

            class ipv6_devconf(nla):
                # ./include/uapi/linux/ipv6.h
                # DEVCONF_
                fmt = 'I' * 30
                fields = ('forwarding',
                          'hop_limit',
                          'mtu',
                          'accept_ra',
                          'accept_redirects',
                          'autoconf',
                          'dad_transmits',
                          'router_solicitations',
                          'router_solicitation_interval',
                          'router_solicitation_delay',
                          'use_tempaddr',
                          'temp_valid_lft',
                          'temp_prefered_lft',
                          'regen_max_retry',
                          'max_desync_factor',
                          'max_addresses',
                          'force_mld_version',
                          'accept_ra_defrtr',
                          'accept_ra_pinfo',
                          'accept_ra_rtr_pref',
                          'router_probe_interval',
                          'accept_ra_rt_info_max_plen',
                          'proxy_ndp',
                          'optimistic_dad',
                          'accept_source_route',
                          'mc_forwarding',
                          'disable_ipv6',
                          'accept_dad',
                          'force_tllao',
                          'ndisc_notify')

            class ipv6_cache_info(nla):
                # ./include/uapi/linux/if_link.h: struct ifla_cacheinfo
                fmt = 'I' * 4
                fields = ('max_reasm_len',
                          'tstamp',
                          'reachable_time',
                          'retrans_time')

            class ipv6_stats(nla):
                fmt = 'I' * 31
                fields = ('inoctets',
                          'fragcreates',
                          'indiscards',
                          'num',
                          'outoctets',
                          'outnoroutes',
                          'inbcastoctets',
                          'outforwdatagrams',
                          'outpkts',
                          'reasmtimeout',
                          'inhdrerrors',
                          'reasmreqds',
                          'fragfails',
                          'outmcastpkts',
                          'inaddrerrors',
                          'inmcastpkts',
                          'reasmfails',
                          'outdiscards',
                          'outbcastoctets',
                          'inmcastoctets',
                          'inpkts',
                          'fragoks',
                          'intoobigerrors',
                          'inunknownprotos',
                          'intruncatedpkts',
                          'outbcastpkts',
                          'reasmoks',
                          'inbcastpkts',
                          'innoroutes',
                          'indelivers',
                          'outmcastoctets')

            class icmp6_stats(nla):
                fmt = 'Q' * 5
                fields = ('num',
                          'inerrors',
                          'outmsgs',
                          'outerrors',
                          'inmsgs')




from socket import AF_INET
from socket import AF_INET6
from socket import AF_UNSPEC
from pyroute2.netlink import netlink
from pyroute2.netlink import NLM_F_REQUEST
from pyroute2.netlink import NLM_F_ACK
from pyroute2.netlink import NLM_F_DUMP
from pyroute2.netlink import NLM_F_CREATE
from pyroute2.netlink import NLM_F_EXCL
from pyroute2.netlink import marshal
from pyroute2.netlink.generic import NETLINK_ROUTE
from pyroute2.netlink.rtnl.rtmsg import rtmsg
from pyroute2.netlink.rtnl.ndmsg import ndmsg
from pyroute2.netlink.rtnl.ifinfmsg import ifinfmsg
from pyroute2.netlink.rtnl.ifaddrmsg import ifaddrmsg

from pyroute2.common import map_namespace


##  RTnetlink multicast groups
RTNLGRP_NONE = 0x0
RTNLGRP_LINK = 0x1
RTNLGRP_NOTIFY = 0x2
RTNLGRP_NEIGH = 0x4
RTNLGRP_TC = 0x8
RTNLGRP_IPV4_IFADDR = 0x10
RTNLGRP_IPV4_MROUTE = 0x20
RTNLGRP_IPV4_ROUTE = 0x40
RTNLGRP_IPV4_RULE = 0x80
RTNLGRP_IPV6_IFADDR = 0x100
RTNLGRP_IPV6_MROUTE = 0x200
RTNLGRP_IPV6_ROUTE = 0x400
RTNLGRP_IPV6_IFINFO = 0x800
RTNLGRP_DECnet_IFADDR = 0x1000
RTNLGRP_NOP2 = 0x2000
RTNLGRP_DECnet_ROUTE = 0x4000
RTNLGRP_DECnet_RULE = 0x8000
RTNLGRP_NOP4 = 0x10000
RTNLGRP_IPV6_PREFIX = 0x20000
RTNLGRP_IPV6_RULE = 0x40000

## Types of messages
#RTM_BASE = 16
RTM_NEWLINK = 16
RTM_DELLINK = 17
RTM_GETLINK = 18
RTM_SETLINK = 19
RTM_NEWADDR = 20
RTM_DELADDR = 21
RTM_GETADDR = 22
RTM_NEWROUTE = 24
RTM_DELROUTE = 25
RTM_GETROUTE = 26
RTM_NEWNEIGH = 28
RTM_DELNEIGH = 29
RTM_GETNEIGH = 30
RTM_NEWRULE = 32
RTM_DELRULE = 33
RTM_GETRULE = 34
RTM_NEWQDISC = 36
RTM_DELQDISC = 37
RTM_GETQDISC = 38
RTM_NEWTCLASS = 40
RTM_DELTCLASS = 41
RTM_GETTCLASS = 42
RTM_NEWTFILTER = 44
RTM_DELTFILTER = 45
RTM_GETTFILTER = 46
RTM_NEWACTION = 48
RTM_DELACTION = 49
RTM_GETACTION = 50
RTM_NEWPREFIX = 52
RTM_GETMULTICAST = 58
RTM_GETANYCAST = 62
RTM_NEWNEIGHTBL = 64
RTM_GETNEIGHTBL = 66
RTM_SETNEIGHTBL = 67
(RTM_NAMES, RTM_VALUES) = map_namespace('RTM', globals())

rtypes = {'RTN_UNSPEC': 0,
          'RTN_UNICAST': 1,      # Gateway or direct route
          'RTN_LOCAL': 2,        # Accept locally
          'RTN_BROADCAST': 3,    # Accept locally as broadcast
          #                        send as broadcast
          'RTN_ANYCAST': 4,      # Accept locally as broadcast,
          #                        but send as unicast
          'RTN_MULTICAST': 5,    # Multicast route
          'RTN_BLACKHOLE': 6,    # Drop
          'RTN_UNREACHABLE': 7,  # Destination is unreachable
          'RTN_PROHIBIT': 8,     # Administratively prohibited
          'RTN_THROW': 9,        # Not in this table
          'RTN_NAT': 10,         # Translate this address
          'RTN_XRESOLVE': 11}    # Use external resolver

rtprotos = {'RTPROT_UNSPEC': 0,
            'RTPROT_REDIRECT': 1,  # Route installed by ICMP redirects;
            #                        not used by current IPv4
            'RTPROT_KERNEL': 2,    # Route installed by kernel
            'RTPROT_BOOT': 3,      # Route installed during boot
            'RTPROT_STATIC': 4,    # Route installed by administrator
            # Values of protocol >= RTPROT_STATIC are not
            # interpreted by kernel;
            # keep in sync with iproute2 !
            'RTPROT_GATED': 8,      # gated
            'RTPROT_RA': 9,         # RDISC/ND router advertisements
            'RTPROT_MRT': 10,       # Merit MRT
            'RTPROT_ZEBRA': 11,     # Zebra
            'RTPROT_BIRD': 12,      # BIRD
            'RTPROT_DNROUTED': 13,  # DECnet routing daemon
            'RTPROT_XORP': 14,      # XORP
            'RTPROT_NTK': 15,       # Netsukuku
            'RTPROT_DHCP': 16}      # DHCP client

rtscopes = {'RT_SCOPE_UNIVERSE': 0,
            'RT_SCOPE_SITE': 200,
            'RT_SCOPE_LINK': 253,
            'RT_SCOPE_HOST': 254,
            'RT_SCOPE_NOWHERE': 255}


class marshal_rtnl(marshal):
    msg_map = {RTM_NEWLINK: ifinfmsg,
               RTM_DELLINK: ifinfmsg,
               RTM_NEWADDR: ifaddrmsg,
               RTM_DELADDR: ifaddrmsg,
               RTM_NEWROUTE: rtmsg,
               RTM_DELROUTE: rtmsg,
               RTM_NEWNEIGH: ndmsg,
               RTM_DELNEIGH: ndmsg}

    def fix_message(self, msg):
        try:
            msg['event'] = RTM_VALUES[msg['header']['type']]
        except:
            pass


class iproute(netlink):
    marshal = marshal_rtnl
    family = NETLINK_ROUTE
    groups = RTNLGRP_IPV4_IFADDR |\
        RTNLGRP_IPV6_IFADDR |\
        RTNLGRP_IPV4_ROUTE |\
        RTNLGRP_IPV6_ROUTE |\
        RTNLGRP_NEIGH |\
        RTNLGRP_LINK

    # 8<---------------------------------------------------------------
    #
    # Listing methods
    #
    def get_links(self, links=None, family=AF_UNSPEC):
        '''
        Get network interfaces sepcifications.
        '''
        result = []
        links = links or ['all']
        msg_flags = NLM_F_REQUEST | NLM_F_DUMP
        if type(links) not in (list, tuple, set):
            links = [links]
        for index in links:
            msg = ifinfmsg()
            msg['family'] = family
            if index != 'all':
                msg['index'] = index
                msg_flags = NLM_F_REQUEST
            result.extend(self.nlm_request(msg, RTM_GETLINK, msg_flags))
        return result

    def get_neighbors(self, family=AF_UNSPEC):
        '''
        Retrieve ARP cache records.
        '''
        msg = ndmsg()
        msg['family'] = family
        return self.nlm_request(msg, RTM_GETNEIGH)

    def get_addr(self, family=AF_UNSPEC):
        '''
        Get all addresses.
        '''
        msg = ifaddrmsg()
        msg['family'] = family
        return self.nlm_request(msg, RTM_GETADDR)

    def get_routes(self, family=AF_UNSPEC, table=-1):
        '''
        Get all routes. You can specify the table. There
        are 255 routing classes (tables), and the kernel
        returns all the routes on each request. So the
        routine filters routes from full output. By default,
        table == -1, which means that one should get all
        tables.
        '''
        # moreover, the kernel returns records without
        # RTA_DST, which I don't know how to interpret :)

        msg = rtmsg()
        msg['family'] = family
        # msg['table'] = table  # you can specify the table
                                # here, but the kernel will
                                # ignore this setting
        routes = self.nlm_request(msg, RTM_GETROUTE)
        return [k for k in [i for i in routes if 'attrs' in i]
                if [l for l in k['attrs'] if l[0] == 'RTA_DST'] and
                (k['table'] == table or table == -1)]
    # 8<---------------------------------------------------------------

    # 8<---------------------------------------------------------------
    #
    # Shortcuts
    #
    def addr_add(self, *argv, **kwarg):
        '''
        Add an address to an interface
        '''
        return self.addr('add', *argv, **kwarg)

    def addr_del(self, *argv, **kwarg):
        '''
        Remove an address from an interface
        '''
        return self.addr('delete', *argv, **kwarg)

    def route_add(self, *argv, **kwarg):
        '''
        Create a route
        '''
        return self.route('add', *argv, **kwarg)

    def route_del(self, *argv, **kwarg):
        '''
        Delete a route
        '''
        return self.route('delete', *argv, **kwarg)

    def link_up(self, dev):
        '''
        Switch an interface up
        '''
        self.link('set', dev, state='up')

    def link_down(self, dev):
        '''
        Switch an interface down
        '''
        self.link('set', dev, state='down')

    def link_rename(self, dev, name):
        '''
        Rename an interface
        '''
        self.link('set', dev, state='down')
        self.link('set', dev, ifname=name)
        self.link('set', dev, state='up')

    def link_remove(self, dev):
        '''
        Remove an interface
        '''
        self.link('delete', dev)

    def link_names(self):
        '''
        Lookup all interface names. The routine returns a list of
        tuples (index, name).
        '''
        return self.link_filter('IFLA_IFNAME')

    def link_filter(self, name):
        '''
        Show some NLA values for all interfaces.

        Example:

        ip.link_filter("ifname")
        ip.link_filter("address")

        You can supply NLA name in both forms -- "IFLA_ADDRESS" and
        "address" is the same.

        Return format is a list of tuples (index, nla_value).
        '''
        name = name.upper()
        if not name.startswith('IFLA_'):
            name = 'IFLA_%s' % (name)

        return [(l[0], ([m[1] for m in l[1] if m[0] == name] or [None])[0])
                for l in [(k['index'], k['attrs'])
                for k in [i for i in self.get_links() if 'attrs' in i]]]

    def link_lookup(self, **kwarg):
        '''
        Lookup interface index (indeces) by first level NLA
        value.

        Example:

        ip.link_lookup(address="52:54:00:9d:4e:3d")
        ip.link_lookup(ifname="lo")
        ip.link_lookup(operstate="UP")

        Please note, that link_lookup() returns list, not one
        value.
        '''
        name = kwarg.keys()[0]
        value = kwarg[name]

        name = str(name).upper()
        if not name.startswith('IFLA_'):
            name = 'IFLA_%s' % (name)

        return [k['index'] for k in
                [i for i in self.get_links() if 'attrs' in i] if
                [l for l in k['attrs'] if l[0] == name and l[1] == value]]

    # 8<---------------------------------------------------------------

    # 8<---------------------------------------------------------------
    #
    # General low-level configuration methods
    #
    def link(self, action, interface, **kwarg):
        '''
        Link operations.

        * action -- set, add or delete
        * interface -- device index
        * **kwarg -- keywords, NLA

        Example:

        dev = 62  # interface index
        ip.link("set", dev, state="down")
        ip.link("set", dev, address="00:11:22:33:44:55", name="bala")
        ip.link("set", dev, mtu=1000, txqlen=2000)
        ip.link("set", dev, state="up")

        Keywords "state", "flags" and "mask" are reserved. State can
        be "up" or "down", it is a shortcut:

        state="up":   flags=1, mask=1
        state="down": flags=0, mask=0

        For more flags grep IFF_ in the kernel code, until we write
        human-readable flag resolver.

        Other keywords are from ifinfmsg.nla_map, look into the
        corresponding module. You can use the form "ifname" as well
        as "IFLA_IFNAME" and so on, so that's equal:

        ip.link("set", dev, mtu=1000)
        ip.link("set", dev, IFLA_MTU=1000)

        You can also delete interface with:

        ip.link("delete", dev)
        '''

        actions = {'set': RTM_SETLINK,      # almost all operations
                   'add': RTM_NEWLINK,      # no idea, how to use it :)
                   'delete': RTM_DELLINK}   # remove interface
        action = actions.get(action, action)

        msg_flags = NLM_F_REQUEST | NLM_F_ACK | NLM_F_CREATE | NLM_F_EXCL
        msg = ifinfmsg()
        msg['index'] = interface

        flags = kwarg.pop('flags', 0)
        mask = kwarg.pop('mask', 0) or kwarg.pop('change', 0)

        if 'state' in kwarg:
            mask = 1                  # IFF_UP mask
            if kwarg['state'].lower() == 'up':
                flags = 1             # 0 (down) or 1 (up)
            del kwarg['state']

        msg['flags'] = flags
        msg['change'] = mask

        for key in kwarg:
            nla = key.upper()
            if not nla.startswith('IFLA_'):
                nla = 'IFLA_%s' % (nla)
            msg['attrs'].append((nla, kwarg[key]))

        return self.nlm_request(msg, msg_type=action, msg_flags=msg_flags)

    def addr(self, action, interface, address, mask=24, family=AF_INET):
        '''
        Address operations

        * action -- add, delete
        * interface -- device index
        * address -- IPv4 or IPv6 address
        * mask -- address mask
        * family -- socket.AF_INET for IPv4 or socket.AF_INET6 for IPv6

        Example:

        dev = 62
        ip.addr("add", dev, address="10.0.0.1", mask=24)
        ip.addr("add", dev, address="10.0.0.2", mask=24)
        '''

        actions = {'add': RTM_NEWADDR,
                   'delete': RTM_DELADDR}
        action = actions.get(action, action)

        flags = NLM_F_REQUEST | NLM_F_ACK | NLM_F_CREATE | NLM_F_EXCL
        msg = ifaddrmsg()
        msg['index'] = interface
        msg['family'] = family
        msg['prefixlen'] = mask
        msg['scope'] = 0xfe
        if family == AF_INET:
            msg['attrs'] = (('IFA_LOCAL', address),
                            ('IFA_ADDRESS', address))
        elif family == AF_INET6:
            msg['attrs'] = (('IFA_ADDRESS', address), )
        return self.nlm_request(msg, msg_type=action, msg_flags=flags)

    def route(self, action, prefix, mask, table=254,
              rtype='RTN_UNICAST', rtproto='RTPROT_STATIC',
              rtscope='RT_SCOPE_UNIVERSE', interface=None,
              gateway=None, family=AF_INET):
        '''
        Route operations

        * action -- add, delete
        * prefix -- route prefix
        * mask -- route prefix mask
        * table -- routing table to use (default: 254)
        * rtype -- route type (default: "RTN_UNICAST")
        * rtproto -- routing protocol (default: "RTPROT_STATIC")
        * rtscope -- routing scope (default: "RT_SCOPE_UNIVERSE")
        * interface -- via device
        * gateway -- via address
        * family -- socket.AF_INET (default) or socket.AF_INET6

        Example:

        ip.route("add", prefix="10.0.0.0", mask=24, gateway="192.168.0.1")
        '''

        actions = {'add': RTM_NEWROUTE,
                   'delete': RTM_DELROUTE}
        action = actions.get(action, action)

        flags = NLM_F_REQUEST | NLM_F_ACK | NLM_F_CREATE | NLM_F_EXCL
        msg = rtmsg()
        msg['table'] = table
        msg['family'] = family
        msg['proto'] = rtprotos[rtproto]
        msg['type'] = rtypes[rtype]
        msg['scope'] = rtscopes[rtscope]
        msg['dst_len'] = mask
        msg['attrs'] = [('RTA_DST', prefix),
                        ('RTA_TABLE', table)]
        if interface is not None:
            msg['attrs'].append(('RTA_OIF', interface))
        if gateway is not None:
            msg['attrs'].append(('RTA_GATEWAY', gateway))

        return self.nlm_request(msg, msg_type=action,
                                msg_flags=flags)
    # 8<---------------------------------------------------------------
